//
//  AppDelegate.swift
//  DCMotor
//
//  Created by Phidgets on 2018-02-22.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = DCMotor()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    @IBOutlet weak var brakingStrengthEventBox: NSBox!
    @IBOutlet weak var brakingLabel: NSTextField!
    @IBOutlet weak var backEMFEventBox: NSBox!
    @IBOutlet weak var backEMFLabel: NSTextField!
    @IBOutlet weak var velocityEventBox: NSBox!
    @IBOutlet weak var velocityLabel: NSTextField!
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var backEMFCheck: NSButton!
    @IBOutlet weak var targetVelocitySlider: NSSlider!
    @IBOutlet weak var minTargetVelocityLabel: NSTextField!
    @IBOutlet weak var maxTargetVelocityLabel: NSTextField!
    @IBOutlet weak var targetVelocityLabel: NSTextField!
    @IBOutlet weak var accelerationSlider: NSSlider!
    @IBOutlet weak var minAccelerationLabel: NSTextField!
    @IBOutlet weak var maxAccelerationLabel: NSTextField!
    @IBOutlet weak var accelerationLabel: NSTextField!
    @IBOutlet weak var brakingStrengthBox: NSBox!
    @IBOutlet weak var brakingStrengthSlider: NSSlider!
    @IBOutlet weak var minBrakingStrengthLabel: NSTextField!
    @IBOutlet weak var maxBrakingStrengthLabel: NSTextField!
    @IBOutlet weak var brakingStrengthLabel: NSTextField!
    @IBOutlet weak var extraSettingsBox: NSBox!
    @IBOutlet weak var currentLimitSlider: NSSlider!
    @IBOutlet weak var minCurrentLimitLabel: NSTextField!
    @IBOutlet weak var maxCurrentLimitLabel: NSTextField!
    @IBOutlet weak var currentLimitLabel: NSTextField!
    @IBOutlet weak var currentRegulatorGainSlider: NSSlider!
    @IBOutlet weak var minCurrentRegulatorGainLabel: NSTextField!
    @IBOutlet weak var maxCurrentRegulatorGainLabel: NSTextField!
    @IBOutlet weak var currentRegulatorGainLabel: NSTextField!
    @IBOutlet weak var fanModePopup: NSPopUpButton!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0
    
    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.backEMFChange.addHandler(backemfchange_handler)
            let _ = ch.brakingStrengthChange.addHandler(brakingchange_handler)
            let _ = ch.velocityUpdate.addHandler(velocityupdate_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.backEMFChange.removeAllHandlers()
            ch.brakingStrengthChange.removeAllHandlers()
            ch.velocityUpdate.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! DCMotor
            let deviceID = try attachedDevice.getDeviceID()
            let minTargetVelocity = try attachedDevice.getMaxVelocity() * -1
            let maxTargetVelocity = try attachedDevice.getMaxVelocity()
            let targetVelocity = try attachedDevice.getMinVelocity()
            let minAcceleration = try attachedDevice.getMinAcceleration()
            let maxAcceleration = try attachedDevice.getMaxAcceleration()
            let acceleration = try attachedDevice.getAcceleration()
            var minBrakingStrength:Double?
            var maxBrakingStrength:Double?
            var brakingStrength:Double?
            var minCurrentLimit:Double?
            var maxCurrentLimit:Double?
            var currentLimit:Double?
            var minCurrentRegulatorGain:Double?
            var maxCurrentRegulatorGain:Double?
            var currentRegulatorGain:Double?
            var fanMode:FanMode?
            
            switch(deviceID){
            case DeviceID.PN_1065:
                fallthrough
            case DeviceID.PN_DCC1000:
                minBrakingStrength = 0//try attachedDevice.getMinBrakingStrength()
                maxBrakingStrength = 1//try attachedDevice.getMaxBrakingStrength()
                brakingStrength = 0//try attachedDevice.getMinBrakingStrength()
                if(deviceID != DeviceID.PN_DCC1000){
                    break
                }
                minCurrentLimit = try attachedDevice.getMinCurrentLimit()
                maxCurrentLimit = try attachedDevice.getMaxCurrentLimit()
                currentLimit = try attachedDevice.getCurrentLimit()
                minCurrentRegulatorGain = try attachedDevice.getMinCurrentRegulatorGain()
                maxCurrentRegulatorGain = try attachedDevice.getMaxCurrentRegulatorGain()
                currentRegulatorGain = try attachedDevice.getCurrentRegulatorGain()
                fanMode = try attachedDevice.getFanMode()
                break
            default:
                break
            }


            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.settingsBox.isHidden = false

                self.minTargetVelocityLabel.stringValue = String(format:"%.2f", minTargetVelocity)
                self.maxTargetVelocityLabel.stringValue = String(format:"%.2f", maxTargetVelocity)
                self.targetVelocityLabel.stringValue = String(format:"%.2f", targetVelocity)
                self.targetVelocitySlider.minValue = minTargetVelocity
                self.targetVelocitySlider.maxValue = maxTargetVelocity
                self.targetVelocitySlider.doubleValue = targetVelocity
                
                self.minAccelerationLabel.stringValue = String(format:"%.2f", minAcceleration)
                self.maxAccelerationLabel.stringValue = String(format:"%.2f", maxAcceleration)
                self.accelerationLabel.stringValue = String(format:"%.2f", acceleration)
                self.accelerationSlider.minValue = minAcceleration
                self.accelerationSlider.maxValue = maxAcceleration
                self.accelerationSlider.doubleValue = acceleration
                
                switch(deviceID){
                case DeviceID.PN_1060:
                    fallthrough
                case DeviceID.PN_1064:
                    self.velocityEventBox.isHidden = false
                    
                    break
                case DeviceID.PN_1065:
                    self.backEMFEventBox.isHidden = false
                    self.backEMFCheck.isHidden = false
                    fallthrough
                case DeviceID.PN_DCC1000:
                    self.velocityEventBox.isHidden = false
                    self.brakingStrengthBox.isHidden = false
                    self.minBrakingStrengthLabel.stringValue = String(format:"%.2f", minBrakingStrength!)
                    self.maxBrakingStrengthLabel.stringValue = String(format:"%.2f", maxBrakingStrength!)
                    self.brakingStrengthLabel.stringValue = String(format:"%.2f", brakingStrength!)
                    self.brakingStrengthSlider.minValue = minBrakingStrength!
                    self.brakingStrengthSlider.maxValue = maxBrakingStrength!
                    self.brakingStrengthSlider.doubleValue = brakingStrength!
                    if(deviceID != DeviceID.PN_DCC1000){
                        break
                    }
                    self.extraSettingsBox.isHidden = false
                    self.brakingStrengthEventBox.isHidden = false
                    self.currentLimitSlider.minValue = minCurrentLimit!
                    self.currentLimitSlider.maxValue = maxCurrentLimit!
                    self.currentLimitSlider.doubleValue = currentLimit!
                    self.minCurrentLimitLabel.stringValue = String(format:"%.2fA", minCurrentLimit!)
                    self.maxCurrentLimitLabel.stringValue = String(format:"%.2fA", maxCurrentLimit!)
                    self.currentLimitLabel.stringValue = String(format:"%.2fA", currentLimit!)
                    self.currentRegulatorGainSlider.minValue = minCurrentRegulatorGain!
                    self.currentRegulatorGainSlider.maxValue = maxCurrentRegulatorGain!
                    self.currentRegulatorGainSlider.doubleValue = currentRegulatorGain!
                    self.minCurrentRegulatorGainLabel.stringValue = String(format:"%.2f",minCurrentRegulatorGain!)
                    self.maxCurrentRegulatorGainLabel.stringValue = String(format:"%.2f",maxCurrentRegulatorGain!)
                    self.currentRegulatorGainLabel.stringValue = String(format:"%.2f",currentRegulatorGain!)
                    self.fanModePopup.addItem(withTitle: "On")
                    self.fanModePopup.addItem(withTitle: "Off")
                    self.fanModePopup.addItem(withTitle: "Auto")
                    self.fanModePopup.selectItem(at: fanMode == FanMode.on ? 0 : fanMode == FanMode.off ? 1 :2)
                    break
                default:
                    break
                }
    
                //Adjusting window height
                var heightChange:CGFloat = 0
                if(!self.brakingStrengthBox.isHidden){
                    heightChange += self.brakingStrengthBox.frame.size.height
                }
                if(!self.extraSettingsBox.isHidden){
                    heightChange += self.extraSettingsBox.frame.size.height
                }
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height + heightChange
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.settingsBox.isHidden = true
            self.extraSettingsBox.isHidden = true
            self.brakingStrengthBox.isHidden = true
            self.backEMFEventBox.isHidden = true
            self.velocityEventBox.isHidden = true
            self.brakingStrengthEventBox.isHidden = true
            self.backEMFCheck.isHidden = true

            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func backemfchange_handler(sender: DCMotor, backEMF: Double){
        DispatchQueue.main.async {
            self.backEMFLabel.stringValue = String(format:"%.2f", backEMF)
        }
    }
    
    func velocityupdate_handler(sender: DCMotor, velocity: Double){
        DispatchQueue.main.async {
            self.velocityLabel.stringValue = String(format:"%.2f", velocity)
        }
    }
    
    func brakingchange_handler(sender: DCMotor, brakingStrength: Double){
        DispatchQueue.main.async {
            self.brakingLabel.stringValue = String(format:"%.2f", brakingStrength)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func setTargetVelocity(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetVelocity(slider.doubleValue)
            targetVelocityLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setAcceleration(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setAcceleration(slider.doubleValue)
            accelerationLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setBrakingStrength(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetBrakingStrength(slider.doubleValue)
            brakingStrengthLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setCurrentLimit(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setCurrentLimit(slider.doubleValue)
            currentLimitLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setCurrentRegulatorGain(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setCurrentRegulatorGain(slider.doubleValue)
            currentRegulatorGainLabel.stringValue = String(format:"%.2f", slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setFanMode(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setFanMode(FanMode.on)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setFanMode(FanMode.off)
            }
            else{
                try ch.setFanMode(FanMode.auto)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func stopMotor(_ sender: Any) {
        do{
            try ch.setTargetVelocity(0)
            targetVelocityLabel.stringValue = "0.00"
            targetVelocitySlider.doubleValue = 0
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setBackEMFSensing(_ sender: Any) {
        do{
            let check = sender as! NSButton
            try ch.setBackEMFSensingState(check.state == NSControl.StateValue.on ? true : false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
}

